import hashlib
import os
from typing import Any, List
import dill
from pathlib import Path


class LazyTracker:
    def __init__(self):
        """LazyTracked enables you to compute combined hash of things like files, directories, python objects etc."""
        self._hasher = hashlib.md5()

    def add_directories(self, directories: List[str], chunk_num_blocks=128):
        """Include hash of all files inside directory (including files in subdirectories)

        Args:
            directories (List[str]): List of directories to take files from
            chunk_num_blocks (int, optional): How many chunks to read at once. Defaults to 128.
        """
        files_to_check = []

        for directory in directories:
            files_to_check.extend(list(Path(directory).rglob("*")))

        files_to_check = sorted(files_to_check)

        self.add_files(files_to_check, chunk_num_blocks)

    def add_files(self, filepaths: List[str], chunk_num_blocks=128):
        """Include hash of files

        Args:
            filepaths (List[str]): List of paths to files
            chunk_num_blocks (int, optional): How many chunks to read at once. Defaults to 128.
        """
        for p in filepaths:
            if os.path.exists(p):
                with open(p, "rb") as f:
                    while chunk := f.read(chunk_num_blocks * self._hasher.block_size):
                        self._hasher.update(chunk)
            else:
                self._hasher.update(dill.dumps(None))

    def add_hparams(self, hparams: dict):
        """Add hash of python dictionary. Utility function for storing pickable

        Args:
            hparams (dict): A dictionary to be included
        """
        self.add_picklables([hparams])

    def add_picklables(self, objects: List[Any], recursive: bool=False):
        """Include hash of any picklable python objects (pickable by dill)

        Args:
            objects (List[Any]): List of python objects to hash
            recursive (bool): Wheter to track dependencies of object. Eg.: if function calls another function, if 
                recursive is set to true the changes in inner function are also tracked
        """
        for obj in objects:
            self._hasher.update(dill.dumps(obj, recurse=recursive))

    def hash(self) -> str:
        """Compute hash

        Returns:
            str: Computed checksum of all things tracked
        """

        return self._hasher.hexdigest()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lazytracker']

package_data = \
{'': ['*']}

install_requires = \
['dill>=0.3.5,<0.4.0']

setup_kwargs = {
    'name': 'lazytracker',
    'version': '0.2.0',
    'description': 'Caching of functions with respect to code changes and disk-file changes',
    'long_description': None,
    'author': 'Michal Pogoda',
    'author_email': 'michal.pogoda@bards.ai',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mipo57/lazytracker',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from dataclasses import dataclass
import shelve
import os
from textwrap import wrap
from typing import Callable, List, Optional
from lazytracker.lazytracker import LazyTracker
from dill import Pickler, Unpickler
from functools import wraps

shelve.Pickler = Pickler
shelve.Unpickler = Unpickler


def cached(
    cache_dir: str = ".lazytracker",
    input_files: Optional[List[str]] = None,
    input_dirs: Optional[List[str]] = None,
    output_dirs: Optional[List[str]] = None,
    output_files: Optional[List[str]] = None,
):
    """Function decorator for caching execution

    Args:
        cache_dir (str, optional): Directory where the lazytracker cache information will be stored. Defaults to ".lazytracker".
        input_files (Optional[List[str]], optional): name of function parameters, that are paths to input files. Defaults to None.
        input_dirs (Optional[List[str]], optional): name of function parameters, that are paths to directories with input files. Defaults to None.
        output_dirs (Optional[List[str]], optional): name of function parameters, that are paths to otuput files. Defaults to None.
        output_files (Optional[List[str]], optional): name of function parameters, that are paths to directories with output files. Defaults to None.
    """

    def inner_func(function: Callable):
        @wraps(function)
        def wrapper(*args, **kwargs):
            kwargs.update(dict(zip(function.__code__.co_varnames, args)))

            os.makedirs(f"{cache_dir}", exist_ok=True)
            with shelve.open(f"{cache_dir}/tracked_functions", "c") as db:
                input_tracker = LazyTracker()
                input_tracker.add_picklables([function], recursive=True)
                input_tracker.add_hparams(kwargs)
                if input_dirs is not None:
                    input_dirs_values = [kwargs[input_dir] for input_dir in input_dirs]
                    input_tracker.add_directories(input_dirs_values)
                if input_files is not None:
                    input_files_values = [
                        kwargs[input_file] for input_file in input_files
                    ]
                    input_tracker.add_files(input_files_values)

                test_output_tracker = LazyTracker()
                if output_dirs is not None:
                    output_dirs_values = [
                        kwargs[output_dir] for output_dir in output_dirs
                    ]
                    test_output_tracker.add_directories(output_dirs_values)
                if output_files is not None:
                    output_files_values = [
                        kwargs[output_file] for output_file in output_files
                    ]
                    test_output_tracker.add_files(output_files_values)

                input_hash = input_tracker.hash()
                output_hash = test_output_tracker.hash()

                if input_hash in db and db[input_hash]["hash"] == output_hash:
                    return db[input_hash]["return_value"]
                else:
                    return_value = function(**kwargs)

                    output_tracker = LazyTracker()
                    if output_dirs is not None:
                        output_dirs_values = [
                            kwargs[output_dir] for output_dir in output_dirs
                        ]
                        output_tracker.add_directories(output_dirs_values)
                    if output_files is not None:
                        output_files_values = [
                            kwargs[output_file] for output_file in output_files
                        ]
                        output_tracker.add_files(output_files_values)

                    db[input_hash] = {
                        "return_value": return_value,
                        "hash": output_tracker.hash(),
                    }

                    return return_value

        return wrapper

    return inner_func
